/*
   stub for fsync():
      fsync() is UNIX, not POSIX
*/

/*-------------------------------------------------------------------*/
/*                                                                   */
/* fsync                                                             */
/*                                                                   */
/*-------------------------------------------------------------------*/
/*                                                                   */
/* #include <unistd.h>                                               */
/*                                                                   */
/* int fsync(int fildes);                                            */
/*                                                                   */
/*-------------------------------------------------------------------*/
/*                                                                   */
/* This routines will flush all modified data from a file to disc.   */
/* This is what the fsync function on HP-UX normally does.           */
/*                                                                   */
/*-------------------------------------------------------------------*/
/*                                                                   */
/* RETURN VALUE                                                      */
/* A 0 value is returned on success.  A -1 value indicates an error. */
/*                                                                   */
/*-------------------------------------------------------------------*/
/*                                                                   */
/* Error codes:                                                      */
/* EBADF                                                             */
/* EINVAL                                                            */
/*                                                                   */
/*-------------------------------------------------------------------*/
/* Created:  HFF, February 1992.                                     */
/*                                                                   */
/* Changes:                                                          */
/*   2-92    Helmut Fieres.  Created.                                */
/*   3-8-93  Steve Hoogheem.                                         */
/*           Pulled this function out of a file of other functions   */
/*             and converted it to a byte stream file.  Now using    */
/*             POSIX include files.                                  */
/*                                                                   */
/*-------------------------------------------------------------------*/

#include <unistd.h>      /* specified by fsync() */
#include <errno.h>       /* errno and error constants */
#include <mpe.h>         /* ccode(), CCE */
#include <fcntl.h>       /* _MPE_FILENO() */
#include <sys/types.h>   /* types in struct stat */
#include <sys/stat.h>    /* struct stat, fstat() */

#pragma intrinsic  FCONTROL, HPDEBUG


int
fsync(int fildes)
{
   short dummy;
   short mpe_fd;
   struct stat stat_buffer;

   /*-------------------------------------------------------------------*/
   /* Debugging ...                                                     */
   /*-------------------------------------------------------------------*/
#if DEBUG
   {
      int debug_status;
      HPDEBUG( &debug_status, "#wl 'Stopped for FSYNC system call'#" );
   }
#endif

   /* check filedes is a valid file descriptor */
   if (fstat(fildes, &stat_buffer) < 0)
      /* errno already set to EBADF by fstat() */
      return (-1);

   mpe_fd = _MPE_FILENO(fildes);

   /*-------------------------------------------------------------------*/
   /* The FCONTROL code 2 will complete all outstanding I/O for this    */
   /* file. If the FCONTROL fails, errno is set to EINVAL.              */
   /*-------------------------------------------------------------------*/
   FCONTROL(mpe_fd, 2, &dummy);
   if ( ccode() != CCE )
   {
      errno = EINVAL;
      return (-1);
   }

   return (0);
} /* fsync() */
