#define NOCTERMID /* for the library */

/*
   Implementation of getpass function.

   This program is defined as Program 11.8 in
   Stevens' Advanced Programming in the UNIX Environment, p. 350.
*/


#include <signal.h>      /* sigset_t,
                              sigaddset(), sigemptyset(), sigprocmask() */
#include <stdio.h>       /* EOF, FILE, stdin, ctermid(), fopen(), fputs(),
                              getc(), putc(), setbuf() */
#include <termios.h>     /* struct termios, tcgetattr(), tcsetattr() */

/* -- is this necessary?
#include "ourhdr.h"      /* */

#define MAX_PASS_LEN    8     /* max #chars for user to enter */


char *
getpass(const char *prompt)
{
   static char      buf[MAX_PASS_LEN + 1];   /* null byte at end */
   char             *ptr;
   sigset_t         sig, sigsave;
   struct termios   term, termsave;
   FILE             *fp;
   int              c;

#ifdef NOCTERMID
   fp = stdin;
#else
   if ( (fp = fopen(ctermid(NULL), "r+")) == NULL)
      return (NULL);
#endif
   setbuf(fp, NULL);

   sigemptyset(&sig);   /* block SIGINT & SIGTSTP, save signal mask */
   sigaddset(&sig, SIGINT);
   sigaddset(&sig, SIGTSTP);
   sigprocmask(SIG_BLOCK, &sig, &sigsave);

   tcgetattr(fileno(fp), &termsave);   /* save tty state */
   term = termsave;                    /* structure copy */
   term.c_lflag &= ~(ECHO | ECHOE | ECHOK | ECHONL);
   tcsetattr(fileno(fp), TCSAFLUSH, &term);

   fputs(prompt, fp);

   ptr = buf;
   while ( (c = getc(fp)) != EOF && c != '\n')
   {
      if (ptr < &buf[MAX_PASS_LEN])
         *ptr++ = c;
   }
   *ptr = 0;          /* null terminate */
   putc('\n', fp);    /* we echo a newline */

                      /* restore tty state */
   tcsetattr(fileno(fp), TCSAFLUSH, &termsave);

                      /* restore signal mask */
   sigprocmask(SIG_SETMASK, &sigsave, NULL);

#ifndef NOCTERMID
   fclose(fp);        /* done with /dev/tty */
#endif

   return (buf);
} /* getpass() */
