/*
   stubs for: endpwent(), getpwent(), setpwent()      not part of POSIX.1
*/

/*
   #include <sys/types.h>
   #include <pwd.h>

   struct passwd *getpwent(void);

                       Returns: pointer if OK, NULL on error or end of file

   void setpwent(void);

   void endpwent(void);

            from: Stevens' Advanced Programming in the UNIX Environment, p. 147
*/

/*
   To compile:
   shell/iX> c89 -I../INCLUDE -Wc,+e pwent.c
or
   shell/iX> frombyte pwent.c PWENTC
   shell/iX> callci "ccxl PWENTC,PWENTO,$NULL ;info='-Ap -IINCLUDE +e'"
   shell/iX> mv PWENTO pwent.o
*/

/*
   Since these routines require PM, it is probably best to put them in an XL;
   linking them with the program will require the program to have PM.  
   shell/iX> addxl PWGR.XL ./pwent.o
or
   shell/iX> callci linkedit
      buildxl PWGR.XL
      addxl from=./pwent.o ;to=PWGR.XL ;privlev=2
      exit
*/


/* do NOT define _POSIX_SOURCE here - want UNIX definition of struct passwd */

#undef _POSIX_SOURCE

#include <pwd.h>         /* UNIX struct passwd; use INCLUDE version */

#include <stdio.h>       /* printf() for DEBUG purposes */
#include <string.h>      /* memset(), strchr(), strcmp(), strcpy() */

#define _MPEXL_SOURCE    /* to pull _mpe_errno, _mpe_intrinsic, t_mpe_status
                            out of <errno.h> */
#include <errno.h>       /* as above */

#ifdef PRINT_ERROR
#include "mpeerrno.h"    /* values for _mpe_intrinsic */
#endif


/************************/
/*   MPE information    */
/************************/

#define MPE_ACCOUNT_LEN    16
#define MPE_FILE_LEN       16
#define MPE_GROUP_LEN      16
#define MPE_USER_PW_LEN    16
#define MPE_USER_LEN       16

typedef struct {
   char user[MPE_USER_LEN];
   char group[MPE_GROUP_LEN];
   char account[MPE_ACCOUNT_LEN];
} directory_name_type;

typedef void ^globalanypointer;

#define PATHNAME_LEN   1023
typedef struct {
   int  length;
   char pathname[PATHNAME_LEN+1];
} pathname_type;


/***********************/
/*   AIF information   */
/***********************/

#define AIF_max_items    10
typedef int              itemnum_array_type;
typedef globalanypointer item_array_type;
typedef t_mpe_status     itemstatus_array_type;

#pragma intrinsic_file "AIFINTR.PUB.SYS"
#pragma intrinsic AIFACCTGET, AIFSYSWIDEGET


/***************************/
/*   program information   */
/***************************/

#define USER_NAME_LEN   (MPE_USER_LEN+1+MPE_ACCOUNT_LEN)
#define HOME_DIR_LEN    (1+MPE_ACCOUNT_LEN+1+MPE_GROUP_LEN)
#define LOGON_PROG_LEN  (1+MPE_ACCOUNT_LEN+1+MPE_GROUP_LEN+1+MPE_FILE_LEN)

#define FALSE 0
#define TRUE  1
typedef int BOOLEAN;

static int beginning_of_file=TRUE;

static BOOLEAN user_dir_eq(const directory_name_type *,
                           const directory_name_type *);



static BOOLEAN
user_dir_eq(const directory_name_type *d1, const directory_name_type *d2)
{
   int i;

   for (i=0; i<MPE_USER_LEN; i++)
      if (d1->user[i] != d2->user[i])
         return (FALSE);
   /* group doesn't matter; should be blank */
   for (i=0; i<MPE_ACCOUNT_LEN; i++)
      if (d1->account[i] != d2->account[i])
         return (FALSE);

   return (TRUE);
} /* user_dir_eq() */



void
setpwent(void)
{
   beginning_of_file = TRUE;

#ifdef DEBUG
   printf("setpwent: beginning_of_file=%d\n", beginning_of_file);
#endif
} /* setpwent() */
   


void
endpwent(void)
{
   beginning_of_file = TRUE;

#ifdef DEBUG
   printf("endpwent: beginning_of_file=%d\n", beginning_of_file);
#endif
} /* endpwent() */



struct passwd *
getpwent(void)
{
   /* data to be returned */
   static struct passwd   user;
   static char            user_name[USER_NAME_LEN+1];
   static char            user_password[MPE_USER_PW_LEN+1];
   static char            comment[USER_NAME_LEN+1];
   static char            home_dir[HOME_DIR_LEN+1];
   static char            logon_prog[LOGON_PROG_LEN+1];

   /* save from call to call */
   static directory_name_type   user_search_key;
   static directory_name_type   prev_dir_name;

   t_mpe_status           overall_status;
   int                    one_entry=1;
   itemnum_array_type     itemnum_array[AIF_max_items+1];
   item_array_type        item_array[AIF_max_items];
   itemstatus_array_type  itemstatus_array[AIF_max_items];
   char                   all_users[MPE_USER_LEN+1];
   char                   all_accounts[MPE_ACCOUNT_LEN+1];
   directory_name_type    directory_name;
   int                    uid, gid;
   pathname_type          home_dir_str, logon_prog_str;
   char                   user_home_group[MPE_GROUP_LEN+1];
   int                    i, j;
   char                  *pos_ptr;


   if (beginning_of_file)
   {
      beginning_of_file = FALSE;
#ifdef DEBUG
   printf("getpwent: bof; beginning_of_file=%d\n", beginning_of_file);
#endif

      /* initialize search_key */
      memset(user_search_key.user, ' ', MPE_USER_LEN);
      memset(user_search_key.group, ' ', MPE_GROUP_LEN);
      memset(user_search_key.account, ' ', MPE_ACCOUNT_LEN);

      /* initialize prev_dir_name */
      memset(prev_dir_name.user, ' ', MPE_USER_LEN);
      memset(prev_dir_name.group, ' ', MPE_GROUP_LEN);
      memset(prev_dir_name.account, ' ', MPE_ACCOUNT_LEN);
   }

   /* get next user */
   one_entry = 1;
   itemnum_array[0] = 6001;
   strcpy(all_users,"@               ");
   item_array[0]    = (item_array_type) all_users;
   itemnum_array[1] = 6201;
   strcpy(all_accounts,"@               ");
   item_array[1]    = (item_array_type) all_accounts;
   itemnum_array[2] = 0;
   overall_status.word = itemstatus_array[0].word = 
        itemstatus_array[1].word = 0;
#ifdef WORKAROUND
   GETPRIVMODE();
#endif
   AIFSYSWIDEGET(&overall_status, 6000, NULL, (int *) &directory_name,
                 &one_entry, itemnum_array, item_array, itemstatus_array,
                 (int *) &user_search_key);
#ifdef WORKAROUND
   GETUSERMODE();
#endif
   if (overall_status.word)
   {
      if (overall_status.decode.error_num < 0)
         _mpe_errno = overall_status.word;
      else
      {
         overall_status.decode.error_num--;    /* C is 0 based */
         _mpe_errno = itemstatus_array[overall_status.decode.error_num].word;
      }
#ifdef PRINT_ERROR
      _mpe_intrinsic = INTRINSIC_USES_STATUS;
      pMPEerror("getpwent");
#endif
      return (NULL);
   }

   /* test if end of chain; previous user is retrieved again */
   if (user_dir_eq(&directory_name, &prev_dir_name))
   {
      beginning_of_file = TRUE;
#ifdef DEBUG
      printf("getpwent: eof; beginning_of_file=%d\n", beginning_of_file);
#endif
      return (NULL);
   }

   /* build user_name */
   for (i=0; i<MPE_USER_LEN, directory_name.user[i] != ' '; i++)
      user_name[i] = directory_name.user[i];
   user_name[i] = '.';   i++;
   for (j=0; j<MPE_ACCOUNT_LEN, directory_name.account[j] != ' '; j++, i++)
      user_name[i] = directory_name.account[j];
   user_name[i] = '\0';
   user.pw_name = user_name;
#ifdef DEBUG
   printf("getpwent: pw_name='%s'\n", user.pw_name);
#endif

   /* get other user information */
   itemnum_array[0]         = 6002;
   item_array[0]            = (item_array_type) user_password;
   itemnum_array[1]         = 6011;
   item_array[1]            = (item_array_type) &uid;
   itemnum_array[2]         = 6216;
   item_array[2]            = (item_array_type) &gid;
   itemnum_array[3]         = 6010;
   home_dir_str.length      = PATHNAME_LEN;
   item_array[3]            = (item_array_type) &home_dir_str;
   itemnum_array[4]         = 6012;
   logon_prog_str.length    = PATHNAME_LEN;
   item_array[4]            = (item_array_type) &logon_prog_str;
     /* getting this item for release 4.5 - see below */
   itemnum_array[5]         = 6006;
   item_array[5]            = (item_array_type) user_home_group;
   itemnum_array[6]         = 0;
   overall_status.word = itemstatus_array[0].word =
        itemstatus_array[1].word = itemstatus_array[2].word = 
        itemstatus_array[3].word = itemstatus_array[4].word =
        itemstatus_array[5].word = 0;
#ifdef WORKAROUND
   GETPRIVMODE();
#endif
   AIFACCTGET(&overall_status, itemnum_array, item_array, itemstatus_array,
              &directory_name);
#ifdef WORKAROUND
   GETUSERMODE();
#endif
   if (overall_status.word)
   {
      if (overall_status.decode.error_num < 0)
         _mpe_errno = overall_status.word;
      else
      {
         overall_status.decode.error_num--;    /* C is 0 based */
         _mpe_errno = itemstatus_array[overall_status.decode.error_num].word;
      }
#ifdef PRINT_ERROR
      _mpe_intrinsic = INTRINSIC_USES_STATUS;
      pMPEerror("getpwent");
#endif
      return (NULL);
   }

   /* set up struct pw_passwd with info just retrieved */
   pos_ptr = strchr(user_password, ' ');
   *pos_ptr = '\0';
   user.pw_passwd = user_password;
#ifdef DEBUG
   printf("  getpwent: pw_passwd='%s'\n", user.pw_passwd);
#endif

   user.pw_uid = uid;
#ifdef DEBUG
   printf("  getpwent: pw_uid=%d\n", user.pw_uid);
#endif

   user.pw_gid = gid;
#ifdef DEBUG
   printf("  getpwent: pw_gid=%d\n", user.pw_gid);
#endif

   /* set comment to 'user account' (user_name with '.' replaced with ' ') */
   strcpy(comment, user_name);
   pos_ptr = strchr(comment, '.');
   *pos_ptr = ' ';
   user.pw_gecos = comment;
#ifdef DEBUG
   printf("  getpwent: pw_gecos='%s'\n", user.pw_gecos);
#endif

   for (i=0; i<home_dir_str.length, home_dir_str.pathname[i] != ' '; i++)
      home_dir[i] = home_dir_str.pathname[i];
   home_dir[i] = '\0';
   /*                                                                        */
   /* Operation Notes - 4.5 Release                                          */
   /*                                                                        */
   /* Note that items 6010 and 6012 (Home Directory and Initial Logon        */
   /* Program) will currently return blanks for most users.  This is because */
   /* these fields are currently not updated in the HPUID.PUB.SYS and        */
   /* HPGID.PUB.SYS files for the first release of POSIX.  When this         */
   /* functionality is fully supported in POSIX, then these items will       */
   /* return valid data.  For first release, all users will have /SYS/PUB/CI */
   /* as the initial logon program and all users will have the logon group   */
   /* and account as the home directory.                                     */
   /*                                                                        */
   /*  from: MPE/iX Architected Interface Facitility: OS Ref. Manual, p. 3-5 */
   /*                                                                        */
   if (!strcmp(home_dir,""))
   {
      home_dir[0] = '/';
      for (i=1, j=0; i<MPE_ACCOUNT_LEN,
                     directory_name.account[j] != ' '; j++, i++)
         home_dir[i] = directory_name.account[j];
      home_dir[i] = '/';   i++;
      for (j=0; j<MPE_GROUP_LEN, user_home_group[j] != ' '; j++, i++)
         home_dir[i] = user_home_group[j];
      home_dir[i] = '\0';
   }
   user.pw_dir = home_dir;
#ifdef DEBUG
   printf("  getpwent: pw_dir='%s'\n", user.pw_dir);
#endif

   for (i=0; i<logon_prog_str.length, logon_prog_str.pathname[i] != ' '; i++)
      logon_prog[i] = logon_prog_str.pathname[i];
   logon_prog[i] = '\0';
   if (!strcmp(logon_prog,""))    /* doing this for release 4.5 - see above */
      strcpy(logon_prog,"/SYS/PUB/CI");
   user.pw_shell = logon_prog;
#ifdef DEBUG
   printf("  getpwent: pw_shell='%s'\n", user.pw_shell);
#endif

   /* set other UNIX struct passwd info to null */
   user.pw_age = NULL;
   user.pw_comment = NULL;
   user.pw_audid = 0;
   user.pw_audflg = 0;

   /* returning new user; save as previous for next time */
   prev_dir_name = directory_name;

   return (&user);
} /* getpwent() */
