#include <signal.h>
#include <netdb.h>
#include <sys/types.h>

long sigsetmask(long mask)
{
  sigset_t sigset, rval;

  sigemptyset(&sigset);
  sigset.set[0] = (int) mask;
  sigemptyset(&rval);
  sigprocmask(SIG_SETMASK, &sigset, &rval);
  return (long) rval.set[0];
}

long sigblock(long mask)
{
  sigset_t sigset, rval;

  sigemptyset(&sigset);
  sigset.set[0] = (int) mask;
  sigemptyset(&rval);
  sigprocmask(SIG_BLOCK, &sigset, &rval);
  return (long) rval.set[0];
}

void sigvector(int sig, struct sigvec *vec, struct sigvec *ovec)
{
  sigaction_t act, oact, *actp;

  if (vec) {
    actp = &act;
    act.sa_handler = vec->sv_handler;
    sigemptyset(&act.sa_mask);
    act.sa_mask.set[0] = (int) vec->sv_mask;
    act.sa_flags   = (int) vec->sv_flags;
    }
  else
    actp = NULL;

  sigaction(sig, actp, &oact);

  if (ovec) {
    ovec->sv_handler = oact.sa_handler;
    ovec->sv_mask    = (long) oact.sa_mask.set[0];
    ovec->sv_flags   = (long) oact.sa_flags;
    }
}

long sigpause(long mask)
{
  sigset_t sigset;

  sigemptyset(&sigset);
  sigset.set[0] = (int) mask;
  sigsuspend(&sigset);

  return (long) -1;
}

long sigspace(long size)
{
  return size;
}

void sigvec( int sig, struct sigvec *vec, struct sigvec *ovec)
{
  sigvector(sig, vec, ovec);
}
