/*
   stub for tcgetattr():

   Compile directives:
   DEBUG       - prints information messages.
   PRINT_ERROR - make this function print the error message.
*/

/*
   #include <termios.h>

   int tcgetattr(int filedes, struct termios *termptr);

                                              Returns: 0 if OK, -1 if error

            from: Stevens' Advanced Programming in the UNIX Environment, p. 336
*/

/* 
   Get the terminal attributes of the device associated with fildes and
      stores them at termios_p.  Returns zero if successful, -1 otherwise.

   Error codes:
   EBADF
   ENOTTY

            from: Zlotnick's The POSIX.1 Standard  A Programmer's Guide, p. 301

   ESYSERR
*/
   

/* already declared
#define _POSIX_SOURCE
*/

#include <termios.h>     /* specified by tcgetattr() */
#ifndef VEOR
#define VEOR 11
#endif

#include <stdio.h>       /* stderr, fprintf(), printf() */
#include <mpe.h>         /* ccode(), CCE */
#include <fcntl.h>       /* _MPE_FILENO() */
#include <sys/types.h>   /* struct stat */
#include <sys/stat.h>    /* fstat() */
#include <unistd.h>      /* isatty() */

/* already declared
#define _MPEXL_SOURCE    /* to pull _mpe_errno, _mpe_intrinsic
                            out of <errno.h> */
#include <errno.h>       /* errno & error codes */

#include <mpe.h>    /* values for _mpe_intrinsic */



/***********************/
/*   MPE information   */
/***********************/

#define   TERM_TYPE_STD      1
#define   TERM_TYPE_MODEM    2
#define   TERM_TYPE_VT       5

#pragma intrinsic FDEVICECONTROL, FFILEINFO, XCONTRAP



int
tcgetattr(int filedes, struct termios *tp)
{
#define NOT_SET   -1
   static short  term_type=NOT_SET;

   const short   read_value=1;

   int           ccode_return=0, i=0, oldplabel=0, oldplabel2=0,
                   return_status=0;
   short         fserrorcode=0, mpe_fd=0, value_short=0;
   struct stat   stat_buffer;


#ifdef DEBUG
   printf("\ntcgetattr: start of function\n");
#endif

   /* check filedes is a valid file descriptor */
   if (fstat(filedes, &stat_buffer) < 0)
      /* errno already set to EBADF by fstat() */
      return (-1);

   /* check filedes is associated with a terminal */
   if (isatty(filedes) != 1)
   {
      errno = ENOTTY;
      return (-1);
   }

   mpe_fd = _MPE_FILENO(filedes);


   /* get the terminal type */
   if (term_type == NOT_SET)
   {
      FFILEINFO(mpe_fd, 60, &term_type);
      if ((ccode_return = ccode()) != CCE)
      {
         errno = ESYSERR;
         _mpe_errno = ccode_return;
         _mpe_intrinsic = FFILEINFO_INTRINSIC;
#ifdef PRINT_ERROR
         fprintf(stderr, "tcgetattr: FFILEINFO(60) failed; ccode=%d\n",
                 ccode_return);
#endif
         return (-1);
      }
   }
#ifdef DEBUG
   printf("   term_type = %d\n", term_type);
#endif



   /******************/
   /*  c_iflag bits  */
   /******************/
   tp->c_iflag = 0;



   /* BRKINT */
   /* on HP3000, signal SIGINT is not generated when BREAK is received */
   /* BRKINT flag already set to 0 */



   /* ICRNL */
   /* on HP3000, canonical mode is default and CRs are mapped to NLs */
   tp->c_iflag |= ICRNL;
#ifdef DEBUG
   printf("   ICRNL = %d\n", (tp->c_iflag & ICRNL ? 1 : 0) );
#endif



   /* IGNBRK */
   value_short=0;
   FDEVICECONTROL(mpe_fd, &value_short, -2, 192, 5, read_value, &fserrorcode);
   if ((ccode_return = ccode()) != CCE)
   {
      errno = ESYSERR;
      _mpe_errno = ccode_return;
      _mpe_intrinsic = FDEVICECONTROL_INTRINSIC;
#ifdef PRINT_ERROR
      fprintf(stderr, "%s; ccode=%d, fserrorcode=%d\n",
              "tcgetattr: FDEVICECONTROL(192,5,read) failed",
              ccode_return, fserrorcode);
#endif
      return_status = -1;
   }
   else
   {
#ifdef DEBUG
      printf("   IGNBRK: FDEVICECONTROL(192,5,read) = 0x%02x;", value_short);
#endif
      if (value_short == 0) tp->c_iflag |= IGNBRK;
   }
#ifdef DEBUG
   printf("   IGNBRK = %d\n", (tp->c_iflag & IGNBRK ? 1 : 0) );
#endif



   /* IXON */
   if (term_type != TERM_TYPE_STD && term_type != TERM_TYPE_MODEM)
      tp->c_iflag |= IXON;
   else
   {
      value_short=0;
      FDEVICECONTROL(mpe_fd, &value_short, -2, 192, 26,
                     read_value, &fserrorcode);
      if ((ccode_return = ccode()) != CCE)
      {
         errno = ESYSERR;
         _mpe_errno = ccode_return;
         _mpe_intrinsic = FDEVICECONTROL_INTRINSIC;
#ifdef PRINT_ERROR
         fprintf(stderr, "%s; ccode=%d, fserrorcode=%d\n",
                 "tcgetattr: FDEVICECONTROL(192,26,read) failed",
                 ccode_return, fserrorcode);
#endif
         return_status = -1;
      }
      else
      {
#ifdef DEBUG
         printf("   IXOFF: FDEVICECONTROL(192,26,read) = 0x%02x;", value_short);
#endif
         if (value_short == 1) tp->c_iflag |= IXON;
      }
   }
#ifdef DEBUG
   printf("   IXON = %d\n", (tp->c_iflag & IXON ? 1 : 0) );
#endif



   /******************/
   /*  c_oflag bits  */
   /******************/
   tp->c_oflag = 0;



   /* OPOST */
   tp->c_oflag |= OPOST;     /* default on MPE */
#ifdef DEBUG
   printf("   OPOST = %d\n", (tp->c_oflag & OPOST ? 1 : 0) );
#endif



   /******************/
   /*  c_cflag bits  */
   /******************/
   tp->c_cflag = 0;



   /* CS - character size */
   if (term_type != TERM_TYPE_STD && term_type != TERM_TYPE_MODEM)
      tp->c_cflag |= CS8;
   else
   {
      value_short=0;
      FDEVICECONTROL(mpe_fd, &value_short, -2, 192, 56,
                     read_value, &fserrorcode);
      if ((ccode_return = ccode()) != CCE)
      {
         errno = ESYSERR;
         _mpe_errno = ccode_return;
         _mpe_intrinsic = FDEVICECONTROL_INTRINSIC;
#ifdef PRINT_ERROR
         fprintf(stderr, "%s; ccode=%d, fserrorcode=%d\n",
                 "tcgetattr: FDEVICECONTROL(192,56,read) failed",
                 ccode_return, fserrorcode);
#endif
         return_status = -1;
      }
      else
      {
#ifdef DEBUG
         printf("   CS: FDEVICECONTROL(192,56,read) = 0x%02x;", value_short);
#endif
         if      (value_short == 5)   tp->c_cflag |= CS5;
         else if (value_short == 6)   tp->c_cflag |= CS6;
         else if (value_short == 7)   tp->c_cflag |= CS7;
         else                         tp->c_cflag |= CS8;
      }
   }
#ifdef DEBUG
   {
      int  size;
      char cs_char;

      size = tp->c_cflag & CSIZE;
      if      (size == CS5)   cs_char = '5';
      else if (size == CS6)   cs_char = '6';
      else if (size == CS7)   cs_char = '7';
      else                    cs_char = '8';
      printf("   CS = CS%c\n", cs_char);
   }
#endif



   /* PARENB */
   value_short=0;
   FDEVICECONTROL(mpe_fd, &value_short, -2, 192, 9, read_value, &fserrorcode);
   if ((ccode_return = ccode()) != CCE)
   {
      errno = ESYSERR;
      _mpe_errno = ccode_return;
      _mpe_intrinsic = FDEVICECONTROL_INTRINSIC;
#ifdef PRINT_ERROR
      fprintf(stderr, "%s; ccode=%d, fserrorcode=%d\n",
              "tcgetattr: FDEVICECONTROL(192,9,read) failed",
              ccode_return, fserrorcode);
#endif
      return_status = -1;
   }
   else
   {
#ifdef DEBUG
      printf("   PARENB: FDEVICECONTROL(192,9,read) = 0x%02x;", value_short);
#endif
      if (value_short == 1) tp->c_cflag |= PARENB;
   }
#ifdef DEBUG
   printf("   PARENB = %d\n", (tp->c_cflag & PARENB ? 1 : 0) );
#endif



   /* PARODD */
   value_short=0;
   FDEVICECONTROL(mpe_fd, &value_short, -2, 192, 10, read_value, &fserrorcode);
   if ((ccode_return = ccode()) != CCE)
   {
      errno = ESYSERR;
      _mpe_errno = ccode_return;
      _mpe_intrinsic = FDEVICECONTROL_INTRINSIC;
#ifdef PRINT_ERROR
      fprintf(stderr, "%s; ccode=%d, fserrorcode=%d\n",
              "tcgetattr: FDEVICECONTROL(192,10,read) failed",
              ccode_return, fserrorcode);
#endif
      return_status = -1;
   }
   else
   {
#ifdef DEBUG
      printf("   PARODD: FDEVICECONTROL(192,10,read) = 0x%02x;", value_short);
#endif
      if (value_short == 3) tp->c_cflag |= PARODD;
#ifdef DEBUG
      if (tp->c_cflag & PARENB)
      {
         if (value_short != 3 && value_short != 2)
            fprintf(stderr, "tcgetattr: %s; value = %d\n",
                    "parity type not consistent with parity enabled",
                    value_short);
      }
      else
      {  /* parity disabled */
         if (value_short != 0 && value_short != 1 && value_short != 4)
            fprintf(stderr, "tcgetattr: %s; value = %d\n",
                    "parity type not consistent with parity disabled",
                    value_short);
      }
#endif
   }

#ifdef DEBUG
   printf("   PARODD = %d\n", (tp->c_cflag & PARODD ? 1 : 0) );
#endif



   /* PARFIX */
   /* an HP3000 extension */
   /* how to read and set? */



   /******************/
   /*  c_lflag bits  */
   /******************/
   tp->c_lflag = 0;



   /* ECHO */
   value_short=0;
   FDEVICECONTROL(mpe_fd, &value_short, -2, 192, 4, read_value, &fserrorcode);
   if ((ccode_return = ccode()) != CCE)
   {
      errno = ESYSERR;
      _mpe_errno = ccode_return;
      _mpe_intrinsic = FDEVICECONTROL_INTRINSIC;
#ifdef PRINT_ERROR
      fprintf(stderr, "%s; ccode=%d, fserrorcode=%d\n",
              "tcgetattr: FDEVICECONTROL(192,4,read) failed",
              ccode_return, fserrorcode);
#endif
      return_status = -1;
   }
   else
   {
#ifdef DEBUG
      printf("   ECHO: FDEVICECONTROL(192,4,read) = 0x%02x;", value_short);
#endif
      if (value_short == 1) tp->c_lflag |= ECHO;
   }
#ifdef DEBUG
   printf("   ECHO = %d\n", (tp->c_lflag & ECHO ? 1 : 0) );
#endif



   /* ECHOE */
   if (term_type != TERM_TYPE_STD && term_type != TERM_TYPE_MODEM)
      ; /* ECHOE already set to 0 above */
   else
   {
      value_short=0;
      FDEVICECONTROL(mpe_fd, &value_short, -2, 192, 55,
                     read_value, &fserrorcode);
      if ((ccode_return = ccode()) != CCE)
      {
         errno = ESYSERR;
         _mpe_errno = ccode_return;
         _mpe_intrinsic = FDEVICECONTROL_INTRINSIC;
#ifdef PRINT_ERROR
         fprintf(stderr, "%s; ccode=%d, fserrorcode=%d\n",
                 "tcgetattr: FDEVICECONTROL(192,55,read) failed",
                 ccode_return, fserrorcode);
#endif
         return_status = -1;
      }
      else
      {
#ifdef DEBUG
         printf("   ECHOE: FDEVICECONTROL(192,55,read) = 0x%02x;", value_short);
#endif
         if (value_short == 5) tp->c_lflag |= ECHOE;
      }
   }
#ifdef DEBUG
   printf("   ECHOE = %d\n", (tp->c_lflag & ECHOE ? 1 : 0) );
#endif



   /* ECHOK */
   value_short=0;
   FDEVICECONTROL(mpe_fd, &value_short, -2, 192, 14, read_value, &fserrorcode);
   if ((ccode_return = ccode()) != CCE)
   {
      errno = ESYSERR;
      _mpe_errno = ccode_return;
      _mpe_intrinsic = FDEVICECONTROL_INTRINSIC;
#ifdef PRINT_ERROR
      fprintf(stderr, "%s; ccode=%d, fserrorcode=%d\n",
              "tcgetattr: FDEVICECONTROL(192,14,read) failed",
              ccode_return, fserrorcode);
#endif
      return_status = -1;
   }
   else
   {
#ifdef DEBUG
      printf("   ECHOK: FDEVICECONTROL(192,14,read) = 0x%02x;", value_short);
#endif
      if (value_short == 1) tp->c_lflag |= ECHOK;
   }
#ifdef DEBUG
   printf("   ECHOK = %d\n", (tp->c_lflag & ECHOK ? 1 : 0) );
#endif



   /* ICANON */
   /* any way to read whether line feed is enabled?
      FSETMODE will enable or disable it               */
   tp->c_lflag |= ICANON;    /* default on MPE */
#ifdef DEBUG
   printf("   ICANON = %d\n", (tp->c_lflag & ICANON ? 1 : 0) );
#endif



   /* ISIG */
   oldplabel=0;
   XCONTRAP((int) 0, &oldplabel);
   if ((ccode_return = ccode()) == CCL)
   {
      errno = ESYSERR;
      _mpe_errno = ccode_return;
      _mpe_intrinsic = XCONTRAP_INTRINSIC;
#ifdef PRINT_ERROR
      fprintf(stderr,
              "tcgetattr: XCONTRAP(read) failed; ccode = %d\n", ccode_return);
#endif
      return_status = -1;
   }
   else
   {
#ifdef DEBUG
      printf("   ISIG: oldplabel = %d;", oldplabel);
#endif
      if (oldplabel != 0) tp->c_lflag |= ISIG;

      XCONTRAP((int) oldplabel, &oldplabel2);     /* reset it back */
      if ((ccode_return = ccode()) == CCL)
      {
         errno = ESYSERR;
         _mpe_errno = ccode_return;
         _mpe_intrinsic = XCONTRAP_INTRINSIC;
#ifdef PRINT_ERROR
         fprintf(stderr,
                 "tcgetattr: XCONTRAP(write) failed; ccode = %d\n",
                 ccode_return);
#endif
         return_status = -1;
      }
   }
#ifdef DEBUG
   printf("   ISIG = %d\n", (tp->c_lflag & ISIG ? 1 : 0) );
#endif



   /*********************/
   /*  c_cc characters  */
   /*********************/
   for (i=0; i<NCCS; i++) tp->c_cc[i] = (cc_t) 0;



   /* EOF */
   /* EOF - already set to 0 above */
#ifdef DEBUG
   printf("   EOF = 0x%02x\n", tp->c_cc[VEOF]);
#endif



   /* EOL */
   if (term_type != TERM_TYPE_STD && term_type != TERM_TYPE_MODEM)
      tp->c_cc[VEOL] = (cc_t) 0x0d;     /* ^M */
   else
   {
      value_short=0;
      FDEVICECONTROL(mpe_fd, &value_short, -2, 192, 39,
                     read_value, &fserrorcode);
      if ((ccode_return = ccode()) != CCE)
      {
         errno = ESYSERR;
         _mpe_errno = ccode_return;
         _mpe_intrinsic = FDEVICECONTROL_INTRINSIC;
#ifdef PRINT_ERROR
         fprintf(stderr, "%s; ccode=%d, fserrorcode=%d\n",
                 "tcgetattr: FDEVICECONTROL(192,39,read) failed",
                 ccode_return, fserrorcode);
#endif
         return_status = -1;
      }
      else
      {
#ifdef DEBUG
         printf("   EOL: FDEVICECONTROL(192,39,read) = 0x%04x;", value_short);
#endif
                        /* / 256 to move the value into the high order byte */
         tp->c_cc[VEOL] = (cc_t) (value_short / 256);
      }
   }
#ifdef DEBUG
   printf("   EOL = 0x%02x\n", tp->c_cc[VEOL]);
#endif



   /* ERASE */
   if (term_type != TERM_TYPE_STD && term_type != TERM_TYPE_MODEM)
      tp->c_cc[VERASE] = (cc_t) 0x08;     /* ^H */
   else
   {
      value_short=0;
      FDEVICECONTROL(mpe_fd, &value_short, -2, 192, 36,
                     read_value, &fserrorcode);
      if ((ccode_return = ccode()) != CCE)
      {
         errno = ESYSERR;
         _mpe_errno = ccode_return;
         _mpe_intrinsic = FDEVICECONTROL_INTRINSIC;
#ifdef PRINT_ERROR
         fprintf(stderr, "%s; ccode=%d, fserrorcode=%d\n",
                 "tcgetattr: FDEVICECONTROL(192,36,read) failed",
                 ccode_return, fserrorcode);
#endif
         return_status = -1;
      }
      else
      {
#ifdef DEBUG
         printf("   ERASE: FDEVICECONTROL(192,36,read) = 0x%02x;", value_short);
#endif
         tp->c_cc[VERASE] = (cc_t) value_short;
      }
   }
#ifdef DEBUG
   printf("   ERASE = 0x%02x\n", tp->c_cc[VERASE]);
#endif



   /* INTR */
   if (term_type != TERM_TYPE_STD && term_type != TERM_TYPE_MODEM)
      tp->c_cc[VINTR] = (cc_t) 0x19;     /* ^Y */
   else
   {
      value_short=0;
      FDEVICECONTROL(mpe_fd, &value_short, -2, 192, 41,
                     read_value, &fserrorcode);
      if ((ccode_return = ccode()) != CCE)
      {
         errno = ESYSERR;
         _mpe_errno = ccode_return;
         _mpe_intrinsic = FDEVICECONTROL_INTRINSIC;
#ifdef PRINT_ERROR
         fprintf(stderr, "%s; ccode=%d, fserrorcode=%d\n",
                 "tcgetattr: FDEVICECONTROL(192,41,read) failed",
                 ccode_return, fserrorcode);
#endif
         return_status = -1;
      }
      else
      {
#ifdef DEBUG
         printf("   INTR: FDEVICECONTROL(192,41,read) = 0x%04x;", value_short);
#endif
                        /* / 256 to move the value into the high order byte */
         tp->c_cc[VINTR] = (cc_t) (value_short / 256);
      }
   }
#ifdef DEBUG
   printf("   INTR = 0x%02x\n", tp->c_cc[VINTR]);
#endif



   /* KILL */
   if (term_type != TERM_TYPE_STD && term_type != TERM_TYPE_MODEM)
      tp->c_cc[VKILL] = (cc_t) 0x18;     /* ^X */
   else
   {
      value_short=0;
      FDEVICECONTROL(mpe_fd, &value_short, -2, 192, 37, 
                     read_value, &fserrorcode);
      if ((ccode_return = ccode()) != CCE)
      {
         errno = ESYSERR;
         _mpe_errno = ccode_return;
         _mpe_intrinsic = FDEVICECONTROL_INTRINSIC;
#ifdef PRINT_ERROR
         fprintf(stderr, "%s; ccode=%d, fserrorcode=%d\n",
                 "tcgetattr: FDEVICECONTROL(192,37,read) failed",
                 ccode_return, fserrorcode);
#endif
         return_status = -1;
      }
      else
      {
#ifdef DEBUG
         printf("   KILL: FDEVICECONTROL(192,37,read) = 0x%02x;", value_short);
#endif
         tp->c_cc[VKILL] = (cc_t) value_short;
      }
   }
#ifdef DEBUG
   printf("   KILL = 0x%02x\n", tp->c_cc[VKILL]);
#endif



   /* MIN */
   tp->c_cc[VMIN] = (cc_t) 1;     /* default for MPE */
#ifdef DEBUG
   printf("   MIN = %d\n", tp->c_cc[VMIN]);
#endif



   /* QUIT */
   /* QUIT - already set to 0 above */
   /* On the HP3000, the QUIT character is the BREAK key and
         cannot be reassigned.                                */
#ifdef DEBUG
   printf("   QUIT = 0x%02x\n", tp->c_cc[VQUIT]);
#endif



   /* SUSP */
   /* SUSP - already set to 0 above */
#ifdef DEBUG
   printf("   SUSP = 0x%02x\n", tp->c_cc[VSUSP]);
#endif



   /* START */
   tp->c_cc[VSTART] = (cc_t) 0x11;     /* ^Q */
#ifdef DEBUG
   printf("   START = 0x%02x\n", tp->c_cc[VSTART]);
#endif



   /* STOP */
   tp->c_cc[VSTOP] = (cc_t) 0x13;     /* ^S */
#ifdef DEBUG
   printf("   STOP = 0x%02x\n", tp->c_cc[VSTOP]);
#endif



   /* TIME */
   if (term_type != TERM_TYPE_STD && term_type != TERM_TYPE_MODEM)
      ; /* TIME already set to 0 above */
   else
   {
      value_short=0;
      FDEVICECONTROL(mpe_fd, &value_short, -1, 192, 65,
                     read_value, &fserrorcode);
      if ((ccode_return = ccode()) != CCE)
      {
         if (fserrorcode == 68)
         {
            /* don't error here since not all DTCs have been upgraded */
#ifdef PRINT_ERROR
            fprintf(stderr, "%s; ccode=%d, fserrorcode=%d; %s\n",
                    "tcgetattr: FDEVICECONTROL(192,65,read) failed",
                    ccode_return, fserrorcode,
                    "DTC may need to be upgraded to support interbyte timer");
#endif
            /* TIME - already set to 0 above */
         }
         else
         {
            errno = ESYSERR;
            _mpe_errno = ccode_return;
            _mpe_intrinsic = FDEVICECONTROL_INTRINSIC;
#ifdef PRINT_ERROR
            fprintf(stderr, "%s; ccode=%d, fserrorcode=%d\n",
                    "tcgetattr: FDEVICECONTROL(192,65,read) failed",
                    ccode_return, fserrorcode);
#endif
            return_status = -1;
         }
      }
      else
      {
#ifdef DEBUG
         printf("   TIME: FDEVICECONTROL(192,65,read) = %d;", value_short);
#endif
         tp->c_cc[VTIME] = (cc_t) value_short;   /* both are in 1/10 secs */
      }
   }
#ifdef DEBUG
   printf("   TIME = %d\n", tp->c_cc[VTIME]);
#endif



   /* EOR - an HP3000 extension; same as EOL? */
   if (term_type != TERM_TYPE_STD && term_type != TERM_TYPE_MODEM)
      tp->c_cc[VEOR] = (cc_t) 0x0d;     /* ^M */
   else
   {
      value_short=0;
      FDEVICECONTROL(mpe_fd, &value_short, -2, 192, 39,
                     read_value, &fserrorcode);
      if ((ccode_return = ccode()) != CCE)
      {
         errno = ESYSERR;
         _mpe_errno = ccode_return;
         _mpe_intrinsic = FDEVICECONTROL_INTRINSIC;
#ifdef PRINT_ERROR
         fprintf(stderr, "%s; ccode=%d, fserrorcode=%d\n",
                 "tcgetattr: FDEVICECONTROL(192,39,read) failed",
                 ccode_return, fserrorcode);
#endif
         return_status = -1;
      }
      else
      {
#ifdef DEBUG
         printf("   EOR: FDEVICECONTROL(192,39,read) = 0x%04x;", value_short);
#endif
                        /* / 256 to move the value into the high order byte */
         tp->c_cc[VEOR] = (cc_t) (value_short / 256);
      }
   }
#ifdef DEBUG
   printf("   EOR = 0x%02x\n", tp->c_cc[VEOR]);
#endif



   /**************/
   /*  c_ispeed  */
   /**************/

   tp->c_ispeed = value_short = 0;
   FDEVICECONTROL(mpe_fd, &value_short, -2, 192, 3, read_value, &fserrorcode);
   if ((ccode_return = ccode()) != CCE)
   {
      errno = ESYSERR;
      _mpe_errno = ccode_return;
      _mpe_intrinsic = FDEVICECONTROL_INTRINSIC;
#ifdef PRINT_ERROR
      fprintf(stderr, "%s; ccode=%d, fserrorcode=%d\n",
              "tcgetattr: FDEVICECONTROL(192,3,read) failed",
              ccode_return, fserrorcode);
#endif
      return_status = -1;
   }
   else
   {
#ifdef DEBUG
      printf("   ISPEED: FDEVICECONTROL(192,3,read) = %d;", value_short);
#endif
      tp->c_ispeed = value_short;
   }
#ifdef DEBUG
   printf("   ISPEED = %d\n", tp->c_ispeed);
#endif



   /**************/
   /*  c_ospeed  */
   /**************/

   tp->c_ospeed = tp->c_ispeed;
#ifdef DEBUG
   printf("   OSPEED = %d (same as ISPEED)\n", tp->c_ospeed);
#endif



#ifdef DEBUG
   printf("tcgetattr: end of function\n\n");
#endif
   return (return_status);
} /* tcgetattr() */
