/*
   stub for tcgetpgrp():
*/

/*
   #include <sys/types.h>
   #include <unistd.h>

   pid_t tcgetpgrp(int filedes);

       Returns: process group ID of foreground process group if OK, -1 on error

            from: Stevens' Advanced Programming in the UNIX Environment, p. 248
*/

/*
   Returns the process group ID of the foreground process group asscoc-
      ated with the terminal on success, -1 otherwise.

   Error codes:
   EBADF
   ENOSYS:           if this is not a job control system and does not
                          support the tcgetpgrp() functon.
   ENOTTY

            from: Zlotnick's The POSIX.1 Standard  A Programmer's Guide, p. 301
*/

#define _POSIX_SOURCE

#include <sys/types.h>   /* pid_t, required by getpgrp() & tcgetpgrp() */
#include <unistd.h>      /* _POSIX_JOB_CONTROL (if defined),
                              getpgrp(), isatty(), sysconf(), tcgetpgrp() */
#include <sys/types.h>   /* types in struct stat */
#include <sys/stat.h>    /* struct stat, fstat() */
#include <errno.h>       /* errno & error codes */


/**************************/
/*   global information   */
/**************************/

typedef int BOOLEAN;
#define TRUE   1
#define FALSE  0


pid_t
tcgetpgrp(int filedes)
{
   BOOLEAN       job_control_supported=FALSE;
   long          val=0;
   struct stat   stat_buffer;

#ifdef _POSIX_JOB_CONTROL
   job_control_supported = TRUE;
#else
   errno = 0;
   if ( (val = sysconf(_SC_JOB_CONTROL)) < 0)
   {
      if (errno != 0)
         return (-1);  /* errno has already been set */
      job_control_supported = FALSE;
   }
   else
      job_control_supported = TRUE;
#endif

   if (job_control_supported)
   {
      /* check filedes is a valid file descriptor */
      if (fstat(filedes, &stat_buffer) < 0)
         return (-1);   /* errno already set to EBADF by fstat() */

      /* check filedes is associated with a terminal */
      if (!isatty(filedes))
      {
         errno = ENOTTY;
         return (-1);
      }

      return (getpgrp());
   }
   else
   {
      errno = ENOSYS;   /* job control NOT supported */
      return (-1);
   }
} /* tcgetpgrp() */
