/*
   stub for tcsetattr():

   Compile directives:
   DEBUG       - prints information messages.
   PRINT_ERROR - make this function print the error message.
*/

/*
   #include <termios.h>

   int tcsetattr(int filedes, int opt, const struct termios *termptr);

                                              Returns: 0 if OK, -1 if error

            from: Stevens' Advanced Programming in the UNIX Environment, p. 336
*/

/* 
   Sets the attributes of the terminal device associated with fildes
      according to the contents of the structure at termios_p.  Returns
      zero if partially or completely successful, -1 otherwise.

   Error codes:
   EBADF
   EINVAL
   ENOTTY

            from: Zlotnick's The POSIX.1 Standard  A Programmer's Guide, p. 302

   ESYSERR
*/
   

/* already defined
#define _POSIX_SOURCE */
#define HP3000

                              /* use when struct termios parm is NOT const */
#define STRUCT_IS_NOT_CONST

#include <termios.h>     /* tcsetattr() */
#ifndef VEOR
#define VEOR 11
#endif

#include <stdio.h>       /* stderr, fprintf() */
#include <mpe.h>         /* ccode(), CCE */
#include <fcntl.h>       /* _MPE_FILENO() */
#include <sys/types.h>   /* types in struct stat */
#include <sys/stat.h>    /* struct stat, fstat() */
#include <unistd.h>      /* isatty() */
#include <signal.h>      /* SIGINT, raise() */

/* already defined
#define _MPEXL_SOURCE    /* to pull _mpe_errno, _mpe_intrinsic
                            out of <errno.h> */
#include <errno.h>       /* errno & error codes */

/* part of mpe.h now
#include "mpeerrno.h"    /* values for _mpe_intrinsic */



/***********************/
/*   MPE information   */
/***********************/

#define   TERM_TYPE_STD      1
#define   TERM_TYPE_MODEM    2
#define   TERM_TYPE_VT       5

#pragma intrinsic FCHECK, FCONTROL, FDEVICECONTROL, FFILEINFO, FSETMODE
#pragma intrinsic RESETCONTROL, XCONTRAP



int subsys_break_handler(void);



int
#ifdef STRUCT_IS_NOT_CONST
tcsetattr(int filedes, int opt,       struct termios *tp)
#else
tcsetattr(int filedes, int opt, const struct termios *tp)
#endif
{
#define NOT_SET   -1
   static short  term_type=NOT_SET;

   const short   read_value=1;
   const short   write_value=2;

   int           ccode_return=0, oldplabel=0, return_status=0, size=0;
   struct stat   stat_buffer;
   short         dummy_short=0, fcontrol_item=0, fserrorcode=0, mpe_fd=0,
                   value_short=0;


#ifdef DEBUG
   printf("\ntcsetattr: start of function\n");
#endif

   /* check filedes is a valid file descriptor */
   if (fstat(filedes, &stat_buffer) < 0)
      /* errno already set to EBADF by fstat() */
      return (-1);

   /* check filedes is associated with a terminal */
   if (isatty(filedes) != 1)
   {
      errno = ENOTTY;
      return (-1);
   }

   /* check opt is a valid option */
   if (opt != TCSANOW && opt != TCSADRAIN && opt != TCSAFLUSH)
   {
      errno = EINVAL;
      return (-1);
   }



   mpe_fd = _MPE_FILENO(filedes);

   /* get the terminal type */
   if (term_type == NOT_SET)
   {
      FFILEINFO(mpe_fd, 60, &term_type);
      if ((ccode_return = ccode()) != CCE)
      {
         errno = ESYSERR;
         _mpe_errno = ccode_return;
         _mpe_intrinsic = FFILEINFO_INTRINSIC;
#ifdef PRINT_ERROR
         fprintf(stderr, "tcsetattr: FFILEINFO(60) failed; ccode=%d\n",
                 ccode_return);
#endif
         return (-1);
      }
   }
#ifdef DEBUG
   printf("   term_type = %d\n", term_type);
#endif



   /* TCSADRAIN or TCSAFLUSH - first transmit all output */
   if ((opt == TCSADRAIN || opt == TCSAFLUSH) && term_type != TERM_TYPE_VT)
   {
#ifdef DEBUG
      printf("   opt = %s;", (opt == TCSADRAIN ? "TCSADRAIN" : "TCSAFLUSH") );
#endif
      value_short=1;
      /*
         62   Quiesce I/O
              This directive ensures that all outstanding nowait I/O operations
              have completed.
      */
      FDEVICECONTROL(mpe_fd, &value_short, -2, 192, 62,
                     write_value, &fserrorcode);
      if ((ccode_return = ccode()) != CCE)
      {
         errno = ESYSERR;
         _mpe_errno = ccode_return;
         _mpe_intrinsic = FDEVICECONTROL_INTRINSIC;
#ifdef PRINT_ERROR
         fprintf(stderr, "%s; ccode=%d, fserrorcode=%d\n",
                 "tcsetattr: FDEVICECONTROL(192,62,write) failed",
                 ccode_return, fserrorcode);
#endif
         return (-1);
      }

#ifdef DEBUG
      value_short=0;
      FDEVICECONTROL(mpe_fd, &value_short, -2, 192, 62,
                     read_value, &fserrorcode);
      if ((ccode_return = ccode()) != CCE)
      {
         errno = ESYSERR;
         _mpe_errno = ccode_return;
         _mpe_intrinsic = FDEVICECONTROL_INTRINSIC;
#ifdef PRINT_ERROR
         fprintf(stderr, "%s; ccode=%d, fserrorcode=%d\n",
                 "tcsetattr: FDEVICECONTROL(192,62,read) failed",
                 ccode_return, fserrorcode);
#endif
         return (-1);
      }
      printf("   FDEVICECONTROL(192,62,read) = %d\n", value_short);
#endif /* DEBUG */
   }



   /* TCSAFLUSH - flush all input */
   if (opt == TCSAFLUSH && term_type != TERM_TYPE_VT)
   {
#ifdef DEBUG
      printf("   opt = TCSAFLUSH;");
#endif
      /* if typeahead enabled, flush typeahead buffer */
      value_short=0;
      FDEVICECONTROL(mpe_fd, &value_short, -2, 192, 51,
                     read_value, &fserrorcode);
      if ((ccode_return = ccode()) != CCE)
      {
         errno = ESYSERR;
         _mpe_errno = ccode_return;
         _mpe_intrinsic = FDEVICECONTROL_INTRINSIC;
#ifdef PRINT_ERROR
         fprintf(stderr, "%s; ccode=%d, fserrorcode=%d\n",
                 "tcsetattr: FDEVICECONTROL(192,51,read) failed",
                 ccode_return, fserrorcode);
#endif
         return (-1);
      }
#ifdef DEBUG
      printf("   FDEVICECONTROL(192,51,read) = %d\n", value_short);
#endif

      if (value_short == 1) /* typeahead enabled ? */
      {
         value_short=1;     /* flush typeahead buffer */
         FDEVICECONTROL(mpe_fd, &value_short, -2, 192, 60,
                        write_value, &fserrorcode);
         if ((ccode_return = ccode()) != CCE)
         {
            errno = ESYSERR;
            _mpe_errno = ccode_return;
            _mpe_intrinsic = FDEVICECONTROL_INTRINSIC;
#ifdef PRINT_ERROR
            fprintf(stderr, "%s; ccode=%d, fserrorcode=%d\n",
                    "tcsetattr: FDEVICECONTROL(192,60,write) failed",
                    ccode_return, fserrorcode);
#endif
            return (-1);
         }

#ifdef DEBUG
         value_short=0;
         FDEVICECONTROL(mpe_fd, &value_short, -2, 192, 60,
                        read_value, &fserrorcode);
         if ((ccode_return = ccode()) != CCE)
         {
            errno = ESYSERR;
            _mpe_errno = ccode_return;
            _mpe_intrinsic = FDEVICECONTROL_INTRINSIC;
#ifdef PRINT_ERROR
            fprintf(stderr, "%s; ccode=%d, fserrorcode=%d\n",
                    "tcsetattr: FDEVICECONTROL(192,60,read) failed",
                    ccode_return, fserrorcode);
#endif
            return (-1);
         }
         printf("          FDEVICECONTROL(192,60,read) = %d\n", value_short);
#endif /* DEBUG */
      } /* typeahead enabled ? */
   } /* (opt == TCSAFLUSH) */



   /******************/
   /*  c_iflag bits  */
   /******************/

   /* IGNBRK */
#ifdef DEBUG
   printf("   IGNBRK = %d;", (tp->c_iflag & IGNBRK ? 1 : 0) );
#endif

   fcontrol_item = ( tp->c_iflag & IGNBRK ? 14 : 15 );
   FCONTROL(mpe_fd, fcontrol_item, &dummy_short);
   if ((ccode_return = ccode()) != CCE)
   {
      errno = ESYSERR;
      _mpe_errno = ccode_return;
      _mpe_intrinsic = FCONTROL_INTRINSIC;
#ifdef PRINT_ERROR
      FCHECK(mpe_fd, &fserrorcode);
      fprintf(stderr, "tcsetattr: FCONTROL(%d) failed; ccode=%d; fserror=%d\n",
              fcontrol_item, ccode_return, fserrorcode);
#endif
      return_status = -1;
   }

#ifdef DEBUG
   value_short=0;
   FDEVICECONTROL(mpe_fd, &value_short, -2, 192, 5, read_value, &fserrorcode);
   if ((ccode_return = ccode()) != CCE)
   {
      errno = ESYSERR;
      _mpe_errno = ccode_return;
      _mpe_intrinsic = FDEVICECONTROL_INTRINSIC;
#ifdef PRINT_ERROR
      fprintf(stderr, "%s; ccode=%d, fserrorcode=%d\n",
              "tcsetattr: FDEVICECONTROL(192,5,read) failed",
              ccode_return, fserrorcode);
#endif
      return_status = -1;
   }
   else
      printf("   FDEVICECONTROL(192,5,read) = 0x%02x\n", value_short);
#endif /* DEBUG */



   /* IXON */
   if (term_type == TERM_TYPE_STD || term_type == TERM_TYPE_MODEM)
   {
#ifdef DEBUG
      printf("   IXON = %d;", (tp->c_iflag & IXON ? 1 : 0) );
#endif

      value_short = (tp->c_iflag & IXON ? 1 : 0);
      FDEVICECONTROL(mpe_fd, &value_short, -2, 192, 26,
                     write_value, &fserrorcode);
      if ((ccode_return = ccode()) != CCE)
      {
         errno = ESYSERR;
         _mpe_errno = ccode_return;
         _mpe_intrinsic = FDEVICECONTROL_INTRINSIC;
#ifdef PRINT_ERROR
         fprintf(stderr, "%s; ccode=%d, fserrorcode=%d\n",
                 "tcsetattr: FDEVICECONTROL(192,26,write) failed",
                 ccode_return, fserrorcode);
#endif
         return_status = -1;
      }

#ifdef DEBUG
      value_short=0;
      FDEVICECONTROL(mpe_fd, &value_short, -2, 192, 26,
                     read_value, &fserrorcode);
      if ((ccode_return = ccode()) != CCE)
      {
         errno = ESYSERR;
         _mpe_errno = ccode_return;
         _mpe_intrinsic = FDEVICECONTROL_INTRINSIC;
#ifdef PRINT_ERROR
         fprintf(stderr, "%s; ccode=%d, fserrorcode=%d\n",
                 "tcsetattr: FDEVICECONTROL(192,26,read) failed",
                 ccode_return, fserrorcode);
#endif
         return_status = -1;
      }
      else
         printf("   FDEVICECONTROL(192,26,read) = 0x%02x\n", value_short);
#endif /* DEBUG */
   }



   /******************/
   /*  c_cflag bits  */
   /******************/

   /* CS - character size */
   if (term_type == TERM_TYPE_STD || term_type == TERM_TYPE_MODEM)
   {
      size = tp->c_cflag & CSIZE;
      if      (size == CS5)   value_short = 5;
      else if (size == CS6)   value_short = 6;
      else if (size == CS7)   value_short = 7;
      else                    value_short = 8;
#ifdef DEBUG
      printf("   CS = CS%1d;", value_short);
#endif

      if (size == CS5 || size == CS6)
      {
         errno = ESYSERR;
#ifdef PRINT_ERROR
         fprintf(stderr, "tcsetattr: CS: cannot set to CS%1d on MPE\n",
                 value_short);
#endif
         return_status = -1;
      }
      else
      {
         FDEVICECONTROL(mpe_fd, &value_short, -2, 192, 56,
                        write_value, &fserrorcode);
         if ((ccode_return = ccode()) != CCE)
         {
            errno = ESYSERR;
            _mpe_errno = ccode_return;
            _mpe_intrinsic = FDEVICECONTROL_INTRINSIC;
#ifdef PRINT_ERROR
            fprintf(stderr, "%s; ccode=%d, fserrorcode=%d\n",
                    "tcsetattr: FDEVICECONTROL(192,56,write) failed",
                    ccode_return, fserrorcode);
#endif
            return_status = -1;
         }
      }

#ifdef DEBUG
      value_short=0;
      FDEVICECONTROL(mpe_fd, &value_short, -2, 192, 56,
                     read_value, &fserrorcode);
      if ((ccode_return = ccode()) != CCE)
      {
         errno = ESYSERR;
         _mpe_errno = ccode_return;
         _mpe_intrinsic = FDEVICECONTROL_INTRINSIC;
#ifdef PRINT_ERROR
         fprintf(stderr, "%s; ccode=%d, fserrorcode=%d\n",
                 "tcsetattr: FDEVICECONTROL(192,56,read) failed",
                 ccode_return, fserrorcode);
#endif
         return_status = -1;
      }
      else
         printf("   FDEVICECONTROL(192,56,read) = 0x%02x\n", value_short);
#endif /* DEBUG */
   }



   /* PARENB */
   if (term_type == TERM_TYPE_STD || term_type == TERM_TYPE_MODEM)
   {
#ifdef DEBUG
      printf("   PARENB = %d;", (tp->c_cflag & PARENB ? 1 : 0) );
#endif

      value_short = (tp->c_cflag & PARENB ? 1 : 0);
      FDEVICECONTROL(mpe_fd, &value_short, -2, 192, 9,
                     write_value, &fserrorcode);
      if ((ccode_return = ccode()) != CCE)
      {
         errno = ESYSERR;
         _mpe_errno = ccode_return;
         _mpe_intrinsic = FDEVICECONTROL_INTRINSIC;
#ifdef PRINT_ERROR
         fprintf(stderr, "%s; ccode=%d, fserrorcode=%d\n",
                 "tcsetattr: FDEVICECONTROL(192,9,write) failed",
                 ccode_return, fserrorcode);
#endif
         return_status = -1;
      }

#ifdef DEBUG
      value_short=0;
      FDEVICECONTROL(mpe_fd, &value_short, -2, 192, 9,
                     read_value, &fserrorcode);
      if ((ccode_return = ccode()) != CCE)
      {
         errno = ESYSERR;
         _mpe_errno = ccode_return;
         _mpe_intrinsic = FDEVICECONTROL_INTRINSIC;
#ifdef PRINT_ERROR
         fprintf(stderr, "%s; ccode=%d, fserrorcode=%d\n",
                 "tcsetattr: FDEVICECONTROL(192,9,read) failed",
                 ccode_return, fserrorcode);
#endif
         return_status = -1;
      }
      else
         printf("   FDEVICECONTROL(192,9,read) = 0x%02x\n", value_short);
#endif /* DEBUG */
   }



   /* PARODD */
   if (term_type == TERM_TYPE_STD || term_type == TERM_TYPE_MODEM)
   {
#ifdef DEBUG
      printf("   PARODD = %d;", (tp->c_cflag & PARODD ? 1 : 0) );
#endif

      value_short = (tp->c_cflag & PARODD ? 3 : 2);
      FDEVICECONTROL(mpe_fd, &value_short, -2, 192, 10,
                     write_value, &fserrorcode);
      if ((ccode_return = ccode()) != CCE)
      {
         errno = ESYSERR;
         _mpe_errno = ccode_return;
         _mpe_intrinsic = FDEVICECONTROL_INTRINSIC;
#ifdef PRINT_ERROR
         fprintf(stderr, "%s; ccode=%d, fserrorcode=%d\n",
                 "tcsetattr: FDEVICECONTROL(192,10,write) failed",
                 ccode_return, fserrorcode);
#endif
         return_status = -1;
      }

#ifdef DEBUG
      value_short=0;
      FDEVICECONTROL(mpe_fd, &value_short, -2, 192, 10,
                     read_value, &fserrorcode);
      if ((ccode_return = ccode()) != CCE)
      {
         errno = ESYSERR;
         _mpe_errno = ccode_return;
         _mpe_intrinsic = FDEVICECONTROL_INTRINSIC;
#ifdef PRINT_ERROR
         fprintf(stderr, "%s; ccode=%d, fserrorcode=%d\n",
                 "tcsetattr: FDEVICECONTROL(192,10,read) failed",
                 ccode_return, fserrorcode);
#endif
         return_status = -1;
      }
      else
         printf("   FDEVICECONTROL(192,10,read) = 0x%02x\n", value_short);
#endif /* DEBUG */
   }



   /******************/
   /*  c_lflag bits  */
   /******************/

   /* ECHO */
#ifdef DEBUG
   printf("   ECHO = %d;", (tp->c_lflag & ECHO ? 1 : 0) );
#endif

   fcontrol_item = (tp->c_lflag & ECHO ? 12 : 13);
   value_short = 0;
   FCONTROL(mpe_fd, fcontrol_item, &value_short);
   if ((ccode_return = ccode()) != CCE)
   {
      errno = ESYSERR;
      _mpe_errno = ccode_return;
      _mpe_intrinsic = FCONTROL_INTRINSIC;
#ifdef PRINT_ERROR
      FCHECK(mpe_fd, &fserrorcode);
      fprintf(stderr, "tcsetattr: FCONTROL(%d) failed; ccode=%d; fserror=%d\n",
              fcontrol_item, ccode_return, fserrorcode);
#endif
      return_status = -1;
   }

#ifdef DEBUG
   value_short=0;
   FDEVICECONTROL(mpe_fd, &value_short, -2, 192, 4, read_value, &fserrorcode);
   if ((ccode_return = ccode()) != CCE)
   {
      errno = ESYSERR;
      _mpe_errno = ccode_return;
      _mpe_intrinsic = FDEVICECONTROL_INTRINSIC;
#ifdef PRINT_ERROR
      fprintf(stderr, "%s; ccode=%d, fserrorcode=%d\n",
              "tcsetattr: FDEVICECONTROL(192,4,read) failed",
              ccode_return, fserrorcode);
#endif
      return_status = -1;
   }
   else
      printf("   FDEVICECONTROL(192,4,read) = 0x%02x\n", value_short);
#endif /* DEBUG */



   /* ECHOE */
   if (term_type == TERM_TYPE_STD || term_type == TERM_TYPE_MODEM)
   {
#ifdef DEBUG
      printf("   ECHOE = %d;", (tp->c_lflag & ECHOE ? 1 : 0) );
#endif

      value_short = (tp->c_lflag & ECHOE ? 5 : 1 );
      FDEVICECONTROL(mpe_fd, &value_short,-2, 192, 55,
                     write_value, &fserrorcode);
      if ((ccode_return = ccode()) != CCE)
      {
         errno = ESYSERR;
         _mpe_errno = ccode_return;
         _mpe_intrinsic = FDEVICECONTROL_INTRINSIC;
#ifdef PRINT_ERROR
         fprintf(stderr, "%s; ccode=%d, fserrorcode=%d\n",
                 "tcsetattr: FDEVICECONTROL(192,55,write) failed",
                 ccode_return, fserrorcode);
#endif
         return_status = -1;
      }

#ifdef DEBUG
      value_short=0;
      FDEVICECONTROL(mpe_fd, &value_short, -2, 192, 55,
                     read_value, &fserrorcode);
      if ((ccode_return = ccode()) != CCE)
      {
         errno = ESYSERR;
         _mpe_errno = ccode_return;
         _mpe_intrinsic = FDEVICECONTROL_INTRINSIC;
#ifdef PRINT_ERROR
         fprintf(stderr, "%s; ccode=%d, fserrorcode=%d\n",
                 "tcsetattr: FDEVICECONTROL(192,55,read) failed",
                 ccode_return, fserrorcode);
#endif
         return_status = -1;
      }
      else
         printf("   FDEVICECONTROL(192,55,read) = 0x%02x\n", value_short);
#endif /* DEBUG */
   }



   /* ECHOK */
#ifdef DEBUG
   printf("   ECHOK = %d;", (tp->c_lflag & ECHOK ? 1 : 0) );
#endif

   fcontrol_item = (tp->c_lflag & ECHOK ? 34 : 35 );
   value_short = 0;
   FCONTROL(mpe_fd, fcontrol_item, &value_short);
   if ((ccode_return = ccode()) != CCE)
   {
      errno = ESYSERR;
      _mpe_errno = ccode_return;
      _mpe_intrinsic = FCONTROL_INTRINSIC;
#ifdef PRINT_ERROR
      FCHECK(mpe_fd, &fserrorcode);
      fprintf(stderr, "tcsetattr: FCONTROL(%d) failed; ccode=%d; fserror=%d\n",
              fcontrol_item, ccode_return, fserrorcode);
#endif
      return_status = -1;
   }

#ifdef DEBUG
   value_short=0;
   FDEVICECONTROL(mpe_fd, &value_short, -2, 192, 14, read_value, &fserrorcode);
   if ((ccode_return = ccode()) != CCE)
   {
      errno = ESYSERR;
      _mpe_errno = ccode_return;
      _mpe_intrinsic = FDEVICECONTROL_INTRINSIC;
#ifdef PRINT_ERROR
      fprintf(stderr, "%s; ccode=%d, fserrorcode=%d\n",
              "tcsetattr: FDEVICECONTROL(192,14,read) failed",
              ccode_return, fserrorcode);
#endif
      return_status = -1;
   }
   else
      printf("   FDEVICECONTROL(192,14,read) = 0x%02x\n", value_short);
#endif /* DEBUG */



   /* ICANON */
#ifdef DEBUG
   printf("   ICANON = %d;", (tp->c_lflag & ICANON ? 1 : 0) );
#endif

   /* if canonical, then line feed, else no line feed */
   value_short = ( tp->c_lflag & ICANON ? 0 : 4 );
   FSETMODE(mpe_fd, value_short);
   if ((ccode_return = ccode()) != CCE)
   {
      errno = ESYSERR;
      _mpe_errno = ccode_return;
      _mpe_intrinsic = FSETMODE_INTRINSIC;
#ifdef PRINT_ERROR
      FCHECK(mpe_fd, &fserrorcode);
      fprintf(stderr,
              "tcsetattr: FSETMODE(%d) failed; ccode = %d; fserror=%d\n",
              value_short, ccode_return, fserrorcode);
#endif
      return_status = -1;
   }

#ifdef DEBUG
   /* any way to read whether line feed is enabled? */
   printf("\n");
#endif



   /* ISIG */
#ifdef DEBUG
   printf("   ISIG = %d;", (tp->c_lflag & ISIG ? 1 : 0) );
#endif

   /* ISIG: If set, the INTR, QUIT, and SUSP characters generate signals.
      INTR: On the HP3000, subsystem break is enabled and the subsystem break
               handler generates the INTR signal.
      QUIT: On the HP3000, the QUIT signal cannot be generated.
               System break should be enabled (~IGNBRK) to ensure
               QUIT behavior.  See also QUIT character below.
      SUSP: On the HP3000, the SUSP signal cannot be generated.
               In fact, the suspend character and behavior are not defined. */

   /* ISIG/INTR: The following steps are defined in the Asynchronous Serial
               Communications manual under XCONTRAP, page 8-139 */

   /* 1. define the subsystem break handler - function defined below */

   /* 2. arm or disarm the subsystem break handler */
   if (tp->c_lflag & ISIG)
      XCONTRAP((int) subsys_break_handler, &oldplabel);
   else
      XCONTRAP((int) 0, &oldplabel);
   if ((ccode_return = ccode()) == CCL)
   {
      errno = ESYSERR;
      _mpe_errno = ccode_return;
      _mpe_intrinsic = XCONTRAP_INTRINSIC;
#ifdef PRINT_ERROR
      fprintf(stderr,
              "tcsetattr: XCONTRAP() failed; ccode = %d\n", ccode_return);
#endif
      return_status = -1;
   }

   /* 3. enable or disable subsystem break */
   value_short = (tp->c_lflag & ISIG ? 1 : 0);
   FDEVICECONTROL(mpe_fd, &value_short, -2, 192, 6, write_value, &fserrorcode);
   if ((ccode_return = ccode()) != CCE)
   {
      errno = ESYSERR;
      _mpe_errno = ccode_return;
      _mpe_intrinsic = FDEVICECONTROL_INTRINSIC;
#ifdef PRINT_ERROR
      fprintf(stderr, "%s; ccode=%d, fserrorcode=%d\n",
              "tcsetattr: FDEVICECONTROL(192,6,write) failed",
              ccode_return, fserrorcode);
#endif
      return_status = -1;
   }

#ifdef DEBUG
   value_short=0;
   FDEVICECONTROL(mpe_fd, &value_short, -2, 192, 6, read_value, &fserrorcode);
   if ((ccode_return = ccode()) != CCE)
   {
      errno = ESYSERR;
      _mpe_errno = ccode_return;
      _mpe_intrinsic = FDEVICECONTROL_INTRINSIC;
#ifdef PRINT_ERROR
      fprintf(stderr, "%s; ccode=%d, fserrorcode=%d\n",
              "tcsetattr: FDEVICECONTROL(192,6,read) failed",
              ccode_return, fserrorcode);
#endif
      return_status = -1;
   }
   else
      printf("   FDEVICECONTROL(192,6,read) = 0x%02x\n", value_short);
#endif /* DEBUG */

   /* ISIG/QUIT: System break should be enabled (~IGNBRK). */
   if ((tp->c_lflag & ISIG) && (tp->c_iflag & IGNBRK))
   {
#ifdef PRINT_ERROR
      fprintf(stderr, "tcsetattr: ISIG flag set ,but IGNBRK is set\n");
#endif
   }



   /*********************/
   /*  c_cc characters  */
   /*********************/

   /* EOL */
   if (term_type == TERM_TYPE_STD || term_type == TERM_TYPE_MODEM)
   {
#ifdef DEBUG
      printf("   EOL = 0x%02x;", tp->c_cc[VEOL]);
#endif

      value_short=0;
                        /* * 256 to move the value into the high order byte */
      value_short = ((short) tp->c_cc[VEOL]) * 256;
      FDEVICECONTROL(mpe_fd, &value_short, -2, 192, 39,
                     write_value, &fserrorcode);
      if ((ccode_return = ccode()) != CCE)
      {
         errno = ESYSERR;
         _mpe_errno = ccode_return;
         _mpe_intrinsic = FDEVICECONTROL_INTRINSIC;
#ifdef PRINT_ERROR
         fprintf(stderr, "%s; ccode=%d, fserrorcode=%d\n",
                 "tcsetattr: FDEVICECONTROL(192,39,write) failed",
                 ccode_return, fserrorcode);
#endif
         return_status = -1;
      }

#ifdef DEBUG
      value_short=0;
      FDEVICECONTROL(mpe_fd, &value_short, -2, 192, 39,
                     read_value, &fserrorcode);
      if ((ccode_return = ccode()) != CCE)
      {
         errno = ESYSERR;
         _mpe_errno = ccode_return;
         _mpe_intrinsic = FDEVICECONTROL_INTRINSIC;
#ifdef PRINT_ERROR
         fprintf(stderr, "%s; ccode=%d, fserrorcode=%d\n",
                 "tcsetattr: FDEVICECONTROL(192,39,read) failed",
                 ccode_return, fserrorcode);
#endif
         return_status = -1;
      }
      else
         printf("   FDEVICECONTROL(192,39,read) = 0x%04x\n", value_short);
#endif /* DEBUG */
   }



   /* ERASE */
   if (term_type == TERM_TYPE_STD || term_type == TERM_TYPE_MODEM)
   {
#ifdef DEBUG
      printf("   ERASE = 0x%02x;", tp->c_cc[VERASE]);
#endif

      value_short=0;
      value_short = (short) tp->c_cc[VERASE];
      FDEVICECONTROL(mpe_fd, &value_short, -2, 192, 36,
                     write_value, &fserrorcode);
      if ((ccode_return = ccode()) != CCE)
      {
         errno = ESYSERR;
         _mpe_errno = ccode_return;
         _mpe_intrinsic = FDEVICECONTROL_INTRINSIC;
#ifdef PRINT_ERROR
         fprintf(stderr, "%s; ccode=%d, fserrorcode=%d\n",
                 "tcsetattr: FDEVICECONTROL(192,36,write) failed",
                 ccode_return, fserrorcode);
#endif
         return_status = -1;
      }

#ifdef DEBUG
      value_short=0;
      FDEVICECONTROL(mpe_fd, &value_short, -2, 192, 36,
                     read_value, &fserrorcode);
      if ((ccode_return = ccode()) != CCE)
      {
         errno = ESYSERR;
         _mpe_errno = ccode_return;
         _mpe_intrinsic = FDEVICECONTROL_INTRINSIC;
#ifdef PRINT_ERROR
         fprintf(stderr, "%s; ccode=%d, fserrorcode=%d\n",
                 "tcsetattr: FDEVICECONTROL(192,36,read) failed",
                 ccode_return, fserrorcode);
#endif
         return_status = -1;
      }
      else
         printf("   FDEVICECONTROL(192,36,read) = 0x%02x\n", value_short);
#endif /* DEBUG */
   }



   /* INTR */
   if (term_type == TERM_TYPE_STD || term_type == TERM_TYPE_MODEM)
   {
#ifdef DEBUG
      printf("   INTR = 0x%02x;", tp->c_cc[VINTR]);
#endif

      value_short=0;
                        /* * 256 to move the value into the high order byte */
      value_short = ((short) tp->c_cc[VINTR]) * 256;
      FDEVICECONTROL(mpe_fd, &value_short, -2, 192, 41,
                     write_value, &fserrorcode);
      if ((ccode_return = ccode()) != CCE)
      {
         errno = ESYSERR;
         _mpe_errno = ccode_return;
         _mpe_intrinsic = FDEVICECONTROL_INTRINSIC;
#ifdef PRINT_ERROR
         fprintf(stderr, "%s; ccode=%d, fserrorcode=%d\n",
                 "tcsetattr: FDEVICECONTROL(192,41,write) failed",
                 ccode_return, fserrorcode);
#endif
         return_status = -1;
      }

#ifdef DEBUG
      value_short=0;
      FDEVICECONTROL(mpe_fd, &value_short, -2, 192, 41,
                     read_value, &fserrorcode);
      if ((ccode_return = ccode()) != CCE)
      {
         errno = ESYSERR;
         _mpe_errno = ccode_return;
         _mpe_intrinsic = FDEVICECONTROL_INTRINSIC;
#ifdef PRINT_ERROR
         fprintf(stderr, "%s; ccode=%d, fserrorcode=%d\n",
                 "tcsetattr: FDEVICECONTROL(192,41,read) failed",
                 ccode_return, fserrorcode);
#endif
         return_status = -1;
      }
      else
         printf("   FDEVICECONTROL(192,41,read) = 0x%04x\n", value_short);
#endif /* DEBUG */
   }



   /* KILL */
   if (term_type == TERM_TYPE_STD || term_type == TERM_TYPE_MODEM)
   {
#ifdef DEBUG
      printf("   KILL = 0x%02x;", tp->c_cc[VKILL]);
#endif

      value_short=0;
      value_short = (short) tp->c_cc[VKILL];
      FDEVICECONTROL(mpe_fd, &value_short, -2, 192, 37,
                     write_value, &fserrorcode);
      if ((ccode_return = ccode()) != CCE)
      {
         errno = ESYSERR;
         _mpe_errno = ccode_return;
         _mpe_intrinsic = FDEVICECONTROL_INTRINSIC;
#ifdef PRINT_ERROR
         fprintf(stderr, "%s; ccode=%d, fserrorcode=%d\n",
                 "tcsetattr: FDEVICECONTROL(192,37,write) failed",
                 ccode_return, fserrorcode);
#endif
         return_status = -1;
      }

#ifdef DEBUG
      value_short=0;
      FDEVICECONTROL(mpe_fd, &value_short, -2, 192, 37,
                     read_value, &fserrorcode);
      if ((ccode_return = ccode()) != CCE)
      {
         errno = ESYSERR;
         _mpe_errno = ccode_return;
         _mpe_intrinsic = FDEVICECONTROL_INTRINSIC;
#ifdef PRINT_ERROR
         fprintf(stderr, "%s; ccode=%d, fserrorcode=%d\n",
                 "tcsetattr: FDEVICECONTROL(192,37,read) failed",
                 ccode_return, fserrorcode);
#endif
         return_status = -1;
      }
      else
         printf("   FDEVICECONTROL(192,37,read) = 0x%02x\n", value_short);
#endif /* DEBUG */
   }



   /* MIN - not implemented yet */



   /* QUIT */
   /* On the HP3000, the QUIT character is the BREAK key and
         cannot be reassigned. */



   /* TIME */
   if (term_type == TERM_TYPE_STD || term_type == TERM_TYPE_MODEM)
   {
#ifdef DEBUG
      printf("   TIME = %d;", tp->c_cc[VTIME]);
#endif

      value_short=0;
      value_short = (short) tp->c_cc[VTIME];   /* both are in 1/10 secs */
      FDEVICECONTROL(mpe_fd, &value_short, -2, 192, 65,
                     write_value, &fserrorcode);
      if ((ccode_return = ccode()) != CCE)
      {
         if (fserrorcode == 68)
         {
            /* don't error here since not all DTCs have been upgraded */
#ifdef PRINT_ERROR
            fprintf(stderr, "%s; ccode=%d, fserrorcode=%d; %s\n",
                    "tcsetattr: FDEVICECONTROL(192,65,write) failed",
                    ccode_return, fserrorcode,
                    "DTC may need to be upgraded to support interbyte timer");
#endif
         }
         else
         {
#ifdef DTC_UPGRADE
            errno = ESYSERR;
            _mpe_errno = ccode_return;
            _mpe_intrinsic = FDEVICECONTROL_INTRINSIC;
#ifdef PRINT_ERROR
            fprintf(stderr, "%s; ccode=%d, fserrorcode=%d\n",
                    "tcsetattr: FDEVICECONTROL(192,65,write) failed",
                    ccode_return, fserrorcode);
#endif
            return_status = -1;
#endif /* DTC_UPGRADE */
         }
      }

#ifdef DEBUG
      value_short=0;
      FDEVICECONTROL(mpe_fd, &value_short, -1, 192, 65,
                     read_value, &fserrorcode);
      if ((ccode_return = ccode()) != CCE)
      {
         if (fserrorcode == 68)
         {
            /* don't error here since not all DTCs have been upgraded */
#ifdef PRINT_ERROR
            fprintf(stderr, "%s; ccode=%d, fserrorcode=%d; %s\n",
                    "tcsetattr: FDEVICECONTROL(192,65,read) failed",
                    ccode_return, fserrorcode,
                    "DTC may need to be upgraded to support interbyte timer");
#endif
         }
         else
         {
#ifdef DTC_UPGRADE
            errno = ESYSERR;
            _mpe_errno = ccode_return;
            _mpe_intrinsic = FDEVICECONTROL_INTRINSIC;
#ifdef PRINT_ERROR
            fprintf(stderr, "%s; ccode=%d, fserrorcode=%d\n",
                    "tcsetattr: FDEVICECONTROL(192,65,read) failed",
                    ccode_return, fserrorcode);
#endif
            return_status = -1;
#endif /* DTC_UPGRADE */
         }
      }
      else
         printf("   FDEVICECONTROL(192,65,read) = %d\n", value_short);
#endif /* DEBUG */
   }



   /* EOR - an HP3000 extension; same as EOL? */
   if (term_type == TERM_TYPE_STD || term_type == TERM_TYPE_MODEM)
   {
#ifdef DEBUG
      printf("   EOR = 0x%02x;", tp->c_cc[VEOR]);
#endif

      value_short=0;
                        /* * 256 to move the value into the high order byte */
      value_short = ((short) tp->c_cc[VEOR]) * 256;
      FDEVICECONTROL(mpe_fd, &value_short, -2, 192, 39,
                     write_value, &fserrorcode);
      if ((ccode_return = ccode()) != CCE)
      {
         errno = ESYSERR;
         _mpe_errno = ccode_return;
         _mpe_intrinsic = FDEVICECONTROL_INTRINSIC;
#ifdef PRINT_ERROR
         fprintf(stderr, "%s; ccode=%d, fserrorcode=%d\n",
                 "tcsetattr: FDEVICECONTROL(192,39,write) failed",
                 ccode_return, fserrorcode);
#endif
         return_status = -1;
      }

#ifdef DEBUG
      value_short=0;
      FDEVICECONTROL(mpe_fd, &value_short, -2, 192, 39,
                     read_value, &fserrorcode);
      if ((ccode_return = ccode()) != CCE)
      {
         errno = ESYSERR;
         _mpe_errno = ccode_return;
         _mpe_intrinsic = FDEVICECONTROL_INTRINSIC;
#ifdef PRINT_ERROR
         fprintf(stderr, "%s; ccode=%d, fserrorcode=%d\n",
                 "tcsetattr: FDEVICECONTROL(192,39,read) failed",
                 ccode_return, fserrorcode);
#endif
         return_status = -1;
      }
      else
         printf("   FDEVICECONTROL(192,39,read) = 0x%04x\n", value_short);
#endif /* DEBUG */
   }



   /**************/
   /*  c_ispeed  */
   /**************/

   if (term_type == TERM_TYPE_STD)
   {
#ifdef DEBUG
      printf("   ISPEED = %d;", tp->c_ispeed);
#endif

      fcontrol_item = 10;
      value_short = (short) tp->c_ispeed;
      FCONTROL(mpe_fd, fcontrol_item, &value_short);
      if ((ccode_return = ccode()) != CCE)
      {
         errno = ESYSERR;
         _mpe_errno = ccode_return;
         _mpe_intrinsic = FCONTROL_INTRINSIC;
#ifdef PRINT_ERROR
         FCHECK(mpe_fd, &fserrorcode);
         fprintf(stderr, 
                 "tcsetattr: FCONTROL(%d) failed; ccode=%d, fserror=%d\n",
                 fcontrol_item, ccode_return, fserrorcode);
#endif
         return_status = -1;
      }

#ifdef DEBUG
      value_short=0;
      FDEVICECONTROL(mpe_fd, &value_short, -2, 192, 3,
                     read_value, &fserrorcode);
      if ((ccode_return = ccode()) != CCE)
      {
         errno = ESYSERR;
         _mpe_errno = ccode_return;
         _mpe_intrinsic = FDEVICECONTROL_INTRINSIC;
#ifdef PRINT_ERROR
         fprintf(stderr, "%s; ccode=%d, fserrorcode=%d\n",
                 "tcsetattr: FDEVICECONTROL(192,3,read) failed",
                 ccode_return, fserrorcode);
#endif
         return_status = -1;
      }
      else
         printf("   FDEVICECONTROL(192,3,read) = %d\n", value_short);
#endif /* DEBUG */
   }



   /**************/
   /*  c_ospeed  */
   /**************/

   if (tp->c_ospeed != tp->c_ispeed)
   {
#ifdef STRUCT_IS_NOT_CONST
      tp->c_ospeed = tp->c_ispeed;
#endif

#ifdef DEBUG
      printf(
           "   OSPEED = %d; cannot set different from ISPEED = %d on HP3000\n",
             tp->c_ospeed, tp->c_ispeed);
#endif
   }



#ifdef DEBUG
   printf("tcsetattr: end of function\n\n");
#endif
   return (return_status);
} /* tcsetattr() */



int
subsys_break_handler(void)
{
   int ccode_return=0, return_status=0;

#ifdef DEBUG
   printf("\nin subsys_break_handler()\n");
#endif

   RESETCONTROL();     /* to accept another CONTROL-Y signal */
   if ((ccode_return = ccode()) != CCE)
   {
      errno = ESYSERR;
      _mpe_errno = ccode_return;
      _mpe_intrinsic = RESETCONTROL_INTRINSIC;
#ifdef PRINT_ERROR
      fprintf(stderr, "%s: RESETCONTROL() failed; ccode = %d\n",
              "tcsetattr/subsys_break_handler", ccode_return);
#endif
      return_status = -1;
   }

#ifdef HP3000
   if (kill(0, SIGINT) < 0)
#else
   if (raise(SIGINT) < 0)
#endif
   {
#ifdef PRINT_ERROR
      perror("tcsetattr/subsys_break_handler: kill/raise(SIGINT) failed");
#endif
      return_status = -1;
   }

   return (return_status);
} /* subsys_break_handler() */
