/*
   stub for tcsetpgrp():
*/

/*
   #include <sys/types.h>
   #include <unistd.h>

   int tcsetpgrp(int filedes, pid_t pgrpid);

                                              Returns: 0 if OK, -1 on error

            from: Stevens' Advanced Programming in the UNIX Environment, p. 248
*/

/*
   Sets the foreground process group for the terminal associated with
      filedes to the process group with ID pgrpid.  Returns zero if
      successful, -1 otherwise.

   Error codes:
   EBADF
   EINVAL
   ENOSYS:           if this is not a job control system and does not
                          support the tcsetpgrp() functon.
   ENOTTY
   EPERM

            from: Zlotnick's The POSIX.1 Standard  A Programmer's Guide, p. 302
*/

#define _POSIX_SOURCE

#include <sys/types.h>   /* pid_t, required by tcsetpgrp() */
#include <unistd.h>      /* _POSIX_JOB_CONTROL (if defined),
                              isatty(), sysconf(), tcsetpgrp() */
#include <sys/types.h>   /* types in struct stat */
#include <sys/stat.h>    /* struct stat, fstat() */
#include <errno.h>       /* errno & error codes */


/**************************/
/*   global information   */
/**************************/

typedef int BOOLEAN;
#define TRUE   1
#define FALSE  0


int
tcsetpgrp(int filedes, pid_t pgrpid)
{
   BOOLEAN       job_control_supported=FALSE;
   long          val=0;
   struct stat   stat_buffer;

#ifdef _POSIX_JOB_CONTROL
   job_control_supported = TRUE;
#else
   errno = 0;
   if ( (val = sysconf(_SC_JOB_CONTROL)) < 0)
   {
      if (errno != 0)
         return (-1);  /* errno has already been set */
      job_control_supported = FALSE;
   }
   else
      job_control_supported = TRUE;
#endif

   if (job_control_supported)
   {
      /* check filedes is a valid file descriptor */
      if (fstat(filedes, &stat_buffer) < 0)
         return (-1);   /* errno already set to EBADF by fstat() */

      /* check filedes is associated with a terminal */
      if (!isatty(filedes))
      {
         errno = ENOTTY;
         return (-1);
      }

      return 0;
   }
   else
   {
      errno = ENOSYS;   /* job control NOT supported */
      return (-1);
   }
} /* tcsetpgrp() */
