#ifdef _MPEXL_SOURCE

#define NOCTERMID /* for the library */

/*
   Implementation of getpass function.

   This program is defined as Program 11.8 in
   Stevens' Advanced Programming in the UNIX Environment, p. 350.
*/


#include <signal.h>      /* sigset_t,
                              sigaddset(), sigemptyset(), sigprocmask() */
#include <stdio.h>       /* EOF, FILE, stdin, ctermid(), fopen(), fputs(),
                              getc(), putc(), setbuf() */
#include <termios.h>     /* struct termios, tcgetattr(), tcsetattr() */

/* -- is this necessary?
#include "ourhdr.h"      /* */

#define MAX_PASS_LEN    8     /* max #chars for user to enter */


char *
getpass(const char *prompt)
{
   static char      buf[MAX_PASS_LEN + 1];   /* null byte at end */
   char             *ptr;
   sigset_t         sig, sigsave;
   struct termios   term, termsave;
   FILE             *fp;
   int              c;

#ifdef NOCTERMID
   fp = stdin;
#else
   if ( (fp = fopen(ctermid(NULL), "r+")) == NULL)
      return (NULL);
#endif
   setbuf(fp, NULL);

   sigemptyset(&sig);   /* block SIGINT & SIGTSTP, save signal mask */
   sigaddset(&sig, SIGINT);
   sigaddset(&sig, SIGTSTP);
   sigprocmask(SIG_BLOCK, &sig, &sigsave);

   tcgetattr(fileno(fp), &termsave);   /* save tty state */
   term = termsave;                    /* structure copy */
   term.c_lflag &= ~(ECHO | ECHOE | ECHOK | ECHONL);
   tcsetattr(fileno(fp), TCSAFLUSH, &term);

   fputs(prompt, fp);

   ptr = buf;
   while ( (c = getc(fp)) != EOF && c != '\n')
   {
      if (ptr < &buf[MAX_PASS_LEN])
         *ptr++ = c;
   }
   *ptr = 0;          /* null terminate */
   putc('\n', fp);    /* we echo a newline */

                      /* restore tty state */
   tcsetattr(fileno(fp), TCSAFLUSH, &termsave);

                      /* restore signal mask */
   sigprocmask(SIG_SETMASK, &sigsave, NULL);

#ifndef NOCTERMID
   fclose(fp);        /* done with /dev/tty */
#endif

   return (buf);
} /* getpass() */


#endif /* _MPEXL_SOURCE */



#ifdef VMS
/* Copyright (C) 1991 Free Software Foundation, Inc.
This file is part of the GNU C Library.

The GNU C Library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public License as
published by the Free Software Foundation; either version 2 of the
License, or (at your option) any later version.

The GNU C Library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with the GNU C Library; see the file COPYING.LIB.  If
not, write to the Free Software Foundation, Inc., 675 Mass Ave,
Cambridge, MA 02139, USA.  */

/* CHANGED FOR VMS */

/*
 * <getpass.c>
 */

#include <stdio.h>
#include <descrip.h>
#include <psldef.h>
#include <iodef.h>

char *getpass (const char *prompt)
{
  static char *buf;

  int result;
  $DESCRIPTOR(devnam,"SYS$INPUT");
  int chan;
  int promptlen;
  struct {
     short result;
     short count;
     int   info;
  } iosb;

  promptlen = strlen(prompt);

  buf = malloc(256);
  if (buf == NULL)
     return(NULL);  

  result = SYS$ASSIGN(&devnam, &chan, PSL$C_USER, 0, 0);

  result = SYS$QIOW(0, chan, IO$_READPROMPT | IO$M_PURGE |IO$M_NOECHO, &iosb, 0, 0,
                    buf, 255, 0, 0, prompt, promptlen);

  buf[iosb.count] = '\0';

  result = SYS$DASSGN(chan);

  return buf;
}

#endif /* VMS */
